%---------------------------positive_coupling.m---------------------------%
%
% This file evaluates the performance of flexibly coupled CP (CCP MU - 
% multiplicative update algoritm) against uncoupled CP (MU) for CP models 
% with positive elements in the factors and both measurements and coupling
% follows a Tweedie's distribution (Gamma distribution in the 
% example below).
%
% Two columns of factor A are nearly collinear.
%
% These simulations are related to Table 3 of the paper
% "Exploring multimodal data fusion through joint decompositions with
% flexible couplings".
%
%-------------------------------------------------------------------------%

%--------------Cleaning the workspace and loading functions---------------%
clc;
clear     all;
close     all;
% Add path
addpath('CCP_MU');
%-------------------------------------------------------------------------%

%-----------------------------Model parameters----------------------------%
% 3-way block dimensions
% First tensor
dim       =     [10 10 10];
% Second tensor
dim_p     =     [10 10 dim(3)];
% Number of components
R         =     3;
% Shape parameters for Tweedie's distribution of the measurements
beta_n    =     [2 2];
% Dispersion of the measurements
phi_n     =     [0.5 0.05];
% Shape parameter for the Tweedie's distribution of the coupling
beta_c    =     2;
% Standard deviation of the coupling
phi_c     =     0.05;
% Number of iterations for coupled MU (Multiplicative update)
iter_c    =     20000;
% Number of iterations for uncoupled MU
iter_u    =     5000;
% Number of realizations for the evaluation of the sum of the MSE
n_real    =     200;
% Number of realizations for the initializations
n_init    =     10;
%-------------------------------------------------------------------------%

%------------------------------Tensor generation--------------------------%
% Fixing the seed to have the same tensor
rng(107)
% Random factors generation
A         =     abs(randn(dim(1),R));
B         =     abs(randn(dim(2),R));
A_p       =     abs(randn(dim_p(1),R));
A(:,1)    =     abs(A(:,2)+0.01*randn(dim(1),1));
B_p       =     abs(randn(dim_p(2),R));
C_p       =     abs(randn(dim_p(3),R));

% Normalization
A         =     A.*repmat(1./sum(A),dim(1),1);
B         =     B.*repmat(1./sum(B),dim(2),1);
A_p       =     A_p.*repmat(1./sum(A_p),dim_p(1),1);
B_p       =     B_p.*repmat(1./sum(B_p),dim_p(2),1);

% Noise on the coupling
% Shared factor
C         =     gen_tweedie(C_p,beta_c,phi_c);
%-------------------------------------------------------------------------%

%------------------------Different noise realizations---------------------%
% Progress
N         =     n_real*n_init;
n         =     0;

% Allocation for the squared errors
se_A_u    =     zeros(dim(1)*R,n_real);
se_A_c    =     zeros(dim(1)*R,n_real);
se_A_p_u  =     zeros(dim_p(1)*R,n_real);
se_A_p_c  =     zeros(dim_p(1)*R,n_real);
se_B_u    =     zeros(dim(2)*R,n_real);
se_B_c    =     zeros(dim(2)*R,n_real);
se_B_p_u  =     zeros(dim_p(2)*R,n_real);
se_B_p_c  =     zeros(dim_p(2)*R,n_real);
se_C_u    =     zeros(dim(3)*R,n_real);
se_C_c    =     zeros(dim(3)*R,n_real);
se_C_p_u  =     zeros(dim_p(3)*R,n_real);
se_C_p_c  =     zeros(dim_p(3)*R,n_real);

% Change the seed to have different noise realizations
% rng('shuffle')
for j     =     1:n_real
%Noisy measurements
% First noisy tensor
Y         =     reshape(gen_tweedie(A*transpose(kr(C,B)),...
                                                  beta_n(1),phi_n(1)),dim);
% Second noisy tensor
Y_p       =     reshape(gen_tweedie(A_p*transpose(kr(C_p,B_p)),...
                                                beta_n(2),phi_n(2)),dim_p);

% Different initializations
% Initial cost function values for picking best initialization
obj_n_val_1     =     10^50;
obj_n_val_2     =     10^50;
map_n_val       =     10^50;
for i     =     1:n_init
%-----------------------------Initialization------------------------------%
% Initialization
% Random
A_0         =     abs(randn(size(A)));
B_0         =     abs(randn(size(B)));
A_0         =     A_0.*repmat(1./sum(A_0),dim(1),1);
B_0         =     B_0.*repmat(1./sum(B_0),dim(2),1);
   
A_p_0       =     abs(randn(size(A_p)));
B_p_0       =     abs(randn(size(B_p)));
A_p_0       =     A_p_0.*repmat(1./sum(A_p_0),dim_p(1),1);
B_p_0       =     B_p_0.*repmat(1./sum(B_p_0),dim_p(2),1);
 
C_p_0       =     abs(randn(size(C_p)));
C_0         =     abs(randn(size(C)));
%-------------------------------------------------------------------------%

%----------------------------Uncoupled CP MU------------------------------%
% MU for the first tensor
[A_n_u,B_n_u,C_n_u,obj_val_vec_1]       = ...
                                        mu(Y,beta_n(1),iter_u,A_0,B_0,C_0);
% MU for the second tensor
[A_p_n_u,B_p_n_u,C_p_n_u,obj_val_vec_2] = ...
                                mu(Y_p,beta_n(2),iter_u,A_p_0,B_p_0,C_p_0);
[A_p_n_u,B_p_n_u,C_p_n_u]   =    ...
                             amb_correct_1f_beta(beta_c,A_p_n_u,B_p_n_u,...
                                                            C_p_n_u,C_n_u);                           
if obj_val_vec_1(end)<obj_n_val_1
obj_n_val_1     =     obj_val_vec_1(end);
A_u             =     A_n_u;
B_u             =     B_n_u;
C_u             =     C_n_u;
end
if obj_val_vec_2(end)<obj_n_val_2
obj_n_val_2     =     obj_val_vec_2(end);
A_p_u           =     A_p_n_u;
B_p_u           =     B_p_n_u;
C_p_u           =     C_p_n_u;
end
%-------------------------------------------------------------------------%

%-------------------------Flexibly coupled CP MU--------------------------%
% Initialization with warm start
A_0       =     A_n_u;
B_0       =     B_n_u;
C_0       =     C_n_u;
A_p_0     =     A_p_n_u;
B_p_0     =     B_p_n_u;
C_p_0     =     C_p_n_u;
% CCP_ALS
[A_n_c,B_n_c,C_n_c,A_p_n_c,B_p_n_c,C_p_n_c,map_val_vec] = ...
                          ccp_mu(Y,Y_p,iter_c,beta_n,phi_n,beta_c,phi_c,...
                                            A_0,B_0,C_0,A_p_0,B_p_0,C_p_0);
% Picking best result for different initializations
if map_val_vec(end)<map_n_val
map_n_val       =     map_val_vec(end);
A_c             =     A_n_c;
B_c             =     B_n_c;
C_c             =     C_n_c;
A_p_c           =     A_p_n_c;
B_p_c           =     B_p_n_c;
C_p_c           =     C_p_n_c;
end
%-------------------------------------------------------------------------%
n=n+1;
clc
fprintf('Progress %g %%',(n/N)*100)
end

%--------Correcting the ambiguity and evaluating the squared error--------%
% Uncoupled MU
% Correcting permutation and sign ambiguity for the first tensor
[A_u,B_u,C_u]         =     amb_correct_beta(beta_c,A_u,B_u,C_u,A,B,C);
% Correcting permutation and sign ambiguity for the second tensor
[A_p_u,B_p_u,C_p_u]   =     amb_correct_beta(beta_c,A_p_u,B_p_u,C_p_u,...
                            A_p,B_p,C_p);
% Squared errors on the factors
se_A_u(:,j)     =     (A(:)-A_u(:)).^2;
se_B_u(:,j)     =     (B(:)-B_u(:)).^2;
se_C_u(:,j)     =     (C(:)-C_u(:)).^2;
se_A_p_u(:,j)   =     (A_p(:)-A_p_u(:)).^2;
se_B_p_u(:,j)   =     (B_p(:)-B_p_u(:)).^2;
se_C_p_u(:,j)   =     (C_p(:)-C_p_u(:)).^2;

% Flexibly coupled MU
% Correcting permutation and sign ambiguity for the first tensor
[A_c,B_c,C_c]         =     amb_correct_beta(beta_c,A_c,B_c,C_c,A,B,C);
% Correcting permutation and sign ambiguity for the second tensor
[A_p_c,B_p_c,C_p_c]   =     amb_correct_beta(beta_c,A_p_c,B_p_c,C_p_c,...
                                                              A_p,B_p,C_p);
% Squared errors
se_A_c(:,j)           =     (A(:)-A_c(:)).^2;
se_B_c(:,j)           =     (B(:)-B_c(:)).^2;
se_C_c(:,j)           =     (C(:)-C_c(:)).^2;
se_A_p_c(:,j)         =     (A_p(:)-A_p_c(:)).^2;
se_B_p_c(:,j)         =     (B_p(:)-B_p_c(:)).^2;
se_C_p_c(:,j)         =     (C_p(:)-C_p_c(:)).^2;
%-------------------------------------------------------------------------%
end
%-------------------------------------------------------------------------%

%-------------------------------Results-----------------------------------%
% Mean squared errors on the factors
% Uncoupled MU
mse_A_u         =     sum(mean(se_A_u,2));
mse_B_u         =     sum(mean(se_B_u,2));
mse_C_u         =     sum(mean(se_C_u,2));
mse_A_p_u       =     sum(mean(se_A_p_u,2));
mse_B_p_u       =     sum(mean(se_B_p_u,2));
mse_C_p_u       =     sum(mean(se_C_p_u,2));
% Flexibly coupled MU
mse_A_c         =     sum(mean(se_A_c,2));
mse_B_c         =     sum(mean(se_B_c,2));
mse_C_c         =     sum(mean(se_C_c,2));
mse_A_p_c       =     sum(mean(se_A_p_c,2));
mse_B_p_c       =     sum(mean(se_B_p_c,2));
mse_C_p_c       =     sum(mean(se_C_p_c,2));
%-------------------------------------------------------------------------%

%---------------------------Results Table 3-------------------------------%
fprintf('Alg. \t A \t B \t C \t A_p \t B_p \t C_p\n')
fprintf('Unc. \t %.3f \t %.3f \t %.3f \t %.3f \t %.3f \t %.3f\n',...
                     mse_A_u,mse_B_u,mse_C_u,mse_A_p_u,mse_B_p_u,mse_C_p_u)
fprintf('Coup. \t %.3f \t %.3f \t %.3f \t %.3f \t %.3f \t %.3f\n',...
                     mse_A_c,mse_B_c,mse_C_c,mse_A_p_c,mse_B_p_c,mse_C_p_c)
%-------------------------------------------------------------------------%